open Lwt.Infix

type request =
    | Request of string
    | Answer of string * string

type response =
    | Question of string
    | Body of string
    | Redirect of string
    | Failure of int * string

exception RequestError of string
exception TransportError of string

let ok = Option.get

let authenticator () = match X509.Authenticator.of_string "none" with
    | Ok finalize -> finalize (fun () -> Ptime.of_float_s @@ Unix.time ())
    | Error `Msg msg -> raise @@ TransportError msg

(* The uri has already been sanitized and validated. *)
let hostify uri = uri
    |> Uri.host
    |> ok
    |> Domain_name.of_string_exn
    |> Domain_name.host_exn

let client_config_of uri = 
    let partial = Tls.Config.client ~authenticator: (authenticator ()) in
    try let host = hostify uri in
        partial ~peer_name:host ()
    with Invalid_argument _ ->
        partial ()

let sanitize ?answer uri = Uri.make
    ~scheme: (match Uri.scheme uri with
        | Some scheme ->
            if scheme = "gemini" then scheme
            else raise (RequestError "Invalid scheme provided")
        | None -> "gemini")
    ~host: (match Uri.host uri with
        | Some host -> host
        | None -> raise (RequestError "Host is required"))
    ~port: (match Uri.port uri with
        | Some port -> port
        | None -> 1965)
    ~path: (Uri.path uri)
    ~query: (match answer with
        | Some a -> [(a, [])]
        | None -> Uri.query uri)
    ()

let process_response line body = 
    let parts = Str.bounded_split (Str.regexp " ") line 2 in
    let status = int_of_string @@ List.hd parts in
    let rest = List.nth parts 1 in
    (* Wow do I hate this. *)
    match status with
    | 10 -> Lwt.return @@ Question rest
    | 30 | 31 -> Lwt.return @@ Redirect rest
    | s when s >= 20 && s <= 29 -> Lwt_io.read body
        >>= fun (b) -> Lwt.return @@ Body b
    | s -> Lwt.return @@ Failure (s, rest)

let send_uri_exn uri cfg = Lwt_main.run begin
    match cfg with
    | Ok cfg ->
        Tls_lwt.connect_ext
            cfg
            (ok (Uri.host uri), ok (Uri.port uri))
        >>= fun (ic, oc) ->
            Lwt_io.write oc @@ (Uri.to_string uri) ^ "\r\n"
        >>= fun () ->
            Lwt_io.read_line ic
        >>= fun (resp) -> begin
            process_response resp ic
        end
    | Error `Msg msg -> raise (RequestError msg)
end

let send_exn = function
    | Request ep ->
        let uri = ep
            |> Uri.of_string
            |> sanitize in
        let cfg = client_config_of uri in
        send_uri_exn uri cfg
    | Answer (ep, answer) ->
        let uri = ep
            |> Uri.of_string
            |> sanitize ~answer:answer in
        let cfg = client_config_of uri in
        send_uri_exn uri cfg
